/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 * chapter 01 - exercise 02 - solution 1
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>

#define ALPHA_NUM   10   // Max different letters allowed
#define INT_NUM     10   // Number of integers, in the range [0..9]
#define WORD_LEN (8+1)   // Max word len

typedef struct alpha_s {
  char c;
  int value;
} alpha_t;

/* prototypes */
int index_find (alpha_t *, char, int);
void setup (alpha_t *, char *, int *);
int stringToNumber (alpha_t *, char *, int);
int solution_check (alpha_t *, char *, char *, char *, int);
int solve (alpha_t *, int *, int, char *, char *, char *, int);
void solution_write (alpha_t *, int);
void test (alpha_t *, int *, char *, char *, char *, int);

int main(int argc, char **argv) {
  alpha_t letters[ALPHA_NUM];
  int taken[INT_NUM] = {0};
  int i, letter_used = 0, choice = -1;
  char *w1, *w2, *w3;

  w1 = argv[1];
  w2 = argv[2];
  w3 = argv[3];

  for (i=0; i<ALPHA_NUM; i++) {
    letters[i].value = -1;
    letters[i].c = '\0';
  }
  setup (letters, w1, &letter_used);
  setup (letters, w2, &letter_used);
  setup (letters, w3, &letter_used);

  fprintf (stdout, "(1) Check an existing solution\n");
  fprintf (stdout, "(2) Find a new solution\n");
  fprintf (stdout, "  Action: ");
  fscanf (stdin, "%d", &choice);

  switch(choice) {
    case 1:
      test (letters, taken, w1, w2, w3, letter_used);
      break;
    case 2:
      solve (letters, taken, 0, w1, w2, w3, letter_used);
      break;
    default:
      fprintf (stdout, "Invalid choice\n");
      break;
  }

  return 0;
}

/*
 * Initial set-up
 */
void setup (alpha_t *letters, char *w, int *letter_used) {
  int i;

  for (i=0; i<strlen(w); i++) {
    if (index_find(letters, w[i], *letter_used) == -1)
      letters[(*letter_used)++].c = w[i];
  }

  return;
}

/*
 * Find a given char (toFind) in letters
 */
int index_find (alpha_t *letters, char toFind, int letter_used) {
  int i=0;

  for(i=0; i<letter_used; i++)
    if (letters[i].c == toFind) return i;

  return -1;
}

/*
 * Read and verify the correctness of a given solution
 */
void test (
  alpha_t *letters, int *taken, char *w1, char *w2, char *w3, int letter_used
  )
{
  int i, tmp_val;

  for(i=0; i<letter_used; i++) {
    do {
      fprintf (stdout, "Value for %c? > ", letters[i].c);
      fscanf (stdin, "%d", &tmp_val);
    } while(tmp_val < 0 || tmp_val > INT_NUM || taken[tmp_val]);
    taken[tmp_val] = 1;
    letters[i].value = tmp_val;
  }

  if (solution_check (letters, w1, w2, w3, letter_used)) {
    fprintf (stdout, "The proposed solution is valid!\n");
  } else {
    fprintf (stdout, "The proposed solution does not work!\n");
  }

  return;
}

/*
 * Find a new solution
 */
int solve ( 
  alpha_t *letters, int *taken, int start, char *w1, char *w2,
  char *w3, int letter_used
  )
{
  int solved, i;

  if (start >= letter_used) {
    solved = solution_check (letters, w1, w2, w3, letter_used);
    if (solved != 0)
      solution_write (letters, letter_used);
    return solved;
  }

  for (i=0; i<INT_NUM; i++) {
    if (taken[i]==0) {
      letters[start].value = i;
      taken[i] = 1;
      if (solve (letters, taken, start+1, w1, w2, w3, letter_used))
        return 1;
      letters[start].value = -1;
      taken[i] = 0;
    }
  }

  return 0;
}

/*
 * Verify the correctness of a new solution
 */
int solution_check (
  alpha_t *letters, char *w1, char *w2, char *w3, int letter_used
) {
  int n1, n2, n3;

  n1 = stringToNumber (letters, w1, letter_used);
  n2 = stringToNumber (letters, w2, letter_used);
  n3 = stringToNumber (letters, w3, letter_used);
  if (n1 == -1 || n2 == -1 || n3 == -1) {
    return 0;
  }
  
  return ((n1 + n2) == n3);
}

/*
 * Read a string
 */
void string_read (char *w, FILE *in) {
  fprintf (stdout, "Input string > ");
  fscanf (in, "%s", w);

  return;
}

/*
 * Print-out a solution
 */
void solution_write (alpha_t *letters, int letter_used) {
  int i;

  for(i=0; i<letter_used; i++) {
    fprintf (stdout, "%c > %d\n", letters[i].c, letters[i].value);
  }

  return;
}

/*
 * Convert a word to an integer value
 */
int stringToNumber (alpha_t *letters, char *w, int letter_used) {
  int i, n=0;

  if (letters[index_find (letters, w[0], letter_used)].value == 0) {
    return -1;
  }
  
  for (i=0;i<strlen(w);i++) {
    n = n * 10 +
      (int) (letters[index_find (letters, w[i], letter_used)].value);
  }
  
  return n;
}

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 */
