/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 * chapter 01 - exercise 04 - solution 1
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#define MAX_LEN     (20+1)

// To debug the program set this to TRUE (not zero)
#define DEBUG       1

// Graph edge E
typedef struct edge_s {
  struct vertex_s *vertex;
  int weight;
  struct edge_s *next;
} edge_t;

// Graph vertex V
typedef struct vertex_s {
  char *id;
  int nIn, nOut;
  int visited;
  edge_t *edge;
  struct vertex_s *next;
} vertex_t;

// Graph handle (wrapper)
typedef struct graph_s {
  vertex_t *head;
  int n;
} graph_t;

// Solution structure
typedef struct sol_s {
  vertex_t **path;
  int *weight;
  int len;
  int sum;
} sol_t;

// Prototypes
graph_t *graph_init ();
graph_t *file_read (char *);
vertex_t *vertex_find (graph_t *, char *);
void vertex_add (graph_t *, char *);
void edge_add (graph_t *, vertex_t *, vertex_t *, int);
void graph_write (graph_t *);
void graph_regular (graph_t *);
void path_find (graph_t *, vertex_t *, sol_t *);
void path_find_r (graph_t *, vertex_t *, sol_t *, sol_t *);
void graph_dispose (graph_t *);

int main(int argc, char **argv) {
  graph_t *graph = NULL;
  vertex_t *vertex;
  char id[MAX_LEN];
  sol_t best;

  if (argc != 2) {
    fprintf (stderr, "Run as: pgrmName inputFile\n");
    exit(EXIT_FAILURE);
  }

  graph = file_read (argv[1]);

#if DEBUG
  graph_write (graph);
#endif

  graph_regular (graph);

  best.len = best.sum = (-1);
  best.path = NULL;
  best.weight = NULL;
  do {
    fprintf (stdout, "Insert vertex_t name [end to quit] > ");
    scanf ("%s", id);

    if (strcmp (id, "end") != 0) {
      for (vertex=graph->head; vertex!=NULL; vertex=vertex->next) {
        vertex->visited = 0;
      }
      vertex = vertex_find (graph, id);
      if (vertex == NULL) {
        fprintf (stdout, "Invalid selection\n");
        continue;
      }
      path_find (graph, vertex, &best);
    }
  } while (strcmp (id, "end") != 0);

  free (best.path);
  free (best.weight);
  graph_dispose (graph);

  return 0;
}

/*
 * Read input file
 */
graph_t *file_read (char *name) {
  graph_t *graph;
  vertex_t *vertex1, *vertex2;
  char name1[MAX_LEN], name2[MAX_LEN];
  int weight;
  FILE *fp;

  graph = graph_init ();

  fp = fopen (name, "r");
  if (fp == NULL) {
    fprintf (stderr, "File open error\n");
    exit(EXIT_FAILURE);
  }

  while (fscanf(fp, "%s %d %s", name1, &weight, name2) != EOF) {
    vertex1 = vertex_find (graph, name1);
    if (vertex1 == NULL) {
      vertex_add (graph, name1);
      vertex1 = graph->head;
    }
    vertex2 = vertex_find (graph, name2);
    if (vertex2 == NULL) {
      vertex_add (graph, name2);
      vertex2 = graph->head;
    }
    edge_add (graph, vertex1, vertex2, weight);
  }

  fclose (fp);

  return (graph);
}

/*
 * Allocate graph handle and init fields
 */
graph_t *graph_init () {
  graph_t *g;

  g = (graph_t *) malloc (1 * sizeof(graph_t));
  if (g == NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }
  g->head = NULL;
  g->n = 0;

  return g;
}

/*
 * Find vertex in vertex list
 */
vertex_t *vertex_find (graph_t *graph, char *id) {
  vertex_t *vertex;

  vertex = graph->head;
  while (vertex != NULL) {
    if (strcmp (vertex->id, id) == 0) {
      return vertex;
    }
    vertex = vertex->next;
  }

  return NULL;
}

/*
 * Add a vertex to the graph
 */
void vertex_add (graph_t *graph, char *id) {
  vertex_t *vertex;

  graph->n++;
  vertex = malloc (1 * sizeof (vertex_t));
  if (vertex == NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit (EXIT_FAILURE);
  }
  vertex->id = strdup (id);
  vertex->nIn = 0;
  vertex->nOut = 0;
  vertex->visited = 0;
  vertex->edge = NULL;
  vertex->next = NULL;
  vertex->next = graph->head;
  graph->head = vertex;

  return;
}

/*
 * Add an edge to the graph
 */
void edge_add (
  graph_t *graph, vertex_t *vertex1, vertex_t *vertex2, int weight
) {
  edge_t *edge;
  
  edge = malloc (1 * sizeof(edge_t));
  if (edge == NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }
  edge->vertex = vertex2;
  edge->weight = weight;
  edge->next = vertex1->edge;
  vertex1->edge = edge;
  vertex1->nOut++;
  vertex2->nIn++;
  
  return;
}

/*
 * Print-out graph
 */
void graph_write (graph_t *graph) {
  vertex_t *vertex;
  edge_t *edge;
  int i = 0;

  fprintf (stdout, "GRAPH:\n");
  fprintf (stdout, "Num. of vertices: %d\n", graph->n);
  vertex = graph->head;
  i=0;
  while (vertex != NULL) {
    fprintf (stdout, "#n=%d Vertex=%s inDegree=%d outDegree=%d\n",
      i, vertex->id, vertex->nIn, vertex->nOut);
    edge = vertex->edge;
    while (edge != NULL) {
      fprintf (stdout, "    goes to #%s [weight = %d]\n",
	edge->vertex->id, edge->weight);
      edge = edge->next;
    }
    fprintf (stdout, "\n");

    vertex = vertex->next;
    i++;
  }

  return;
}

/*
 * Check regularity on a graph
 */
void graph_regular (graph_t *graph) {
  vertex_t *vertex;
  int in, out, reg;

  in = out = -1;
  reg = 1;
  for (vertex=graph->head; vertex!=NULL; vertex=vertex->next) {
    if (in == -1 && out == -1) {
      in = vertex->nIn;
      out = vertex->nOut;
      if (in != out) {
        reg = 0;
        break;
      }
    } else {
      if (in != vertex->nIn || out != vertex->nOut) {
        reg = 0;
        break;
      }
    }
  }

  if (reg != 0) {
    fprintf (stdout, "The given graph is regular\n\n");
  } else {
    fprintf (stdout, "The given graph is NOT regular\n\n");
  }

  return;
}

/*
 * Find a path
 */
void path_find (graph_t *graph, vertex_t *vertex, sol_t *best) {
  sol_t curr;
  int i;

  curr.path = (vertex_t **) calloc (graph->n, sizeof (vertex_t *));
  curr.weight = (int *) calloc (graph->n-1, sizeof (int));
  best->path = (vertex_t **) calloc (graph->n, sizeof (vertex_t *));
  best->weight = (int *) calloc (graph->n-1, sizeof (int));
  if (curr.weight==NULL || curr.path==NULL ||
      best->weight==NULL || best->path==NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }
  curr.sum = curr.len = 0;

  path_find_r (graph, vertex, &curr, best);
  
  if (best->sum != -1) {
    for (i=0; i<=best->len; i++) {
      fprintf (stdout, "%s ", best->path[i]->id);
      if (i != best->len)
        fprintf (stdout, "-<w=%d>- ", best->weight[i]);
    }
    fprintf (stdout, ": Total Weight=%d\n", best->sum);
    best->sum = -1;
  } else {
    fprintf (stdout, "No path found...\n");
  }

  free (curr.weight);
  free (curr.path);

  return;
}

/*
 * Find a path (recursive function)
 */
void path_find_r (
  graph_t *graph, vertex_t *vertex, sol_t *curr, sol_t *best)
{
  edge_t *edge;
  int i;

  if (vertex->visited == 1) {
    return;
  }

  curr->path[curr->len] = vertex;
  if (best->sum == -1 || best->sum<curr->sum) {
    for (i=0; i<=curr->len; i++) {
      best->path[i] = curr->path[i];
      if (i != curr->len)
        best->weight[i] = curr->weight[i];
    }
    best->sum = curr->sum;
    best->len = curr->len;
  }

  vertex->visited = 1;

  for (edge=vertex->edge; edge!=NULL; edge=edge->next) {
    curr->weight[curr->len] = edge->weight;
    curr->len = curr->len + 1;
    curr->sum = curr->sum + edge->weight;
    path_find_r (graph, edge->vertex, curr, best);
    curr->len = curr->len - 1;
    curr->sum = curr->sum - edge->weight;
  }

  vertex->visited = 0;

  return;
}

/*
 * Free graph memory
 */
void graph_dispose (graph_t *graph) {
  vertex_t *vertex, *vertexTmp;
  edge_t *edge, *edgeTmp;

  vertex = graph->head;
  while (vertex != NULL) {
    vertexTmp = vertex->next;

    edge = vertex->edge;
    while (edge != NULL) {
      edgeTmp = edge->next;
      free (edge);
      edge = edgeTmp;
    }
    free (vertex);
    vertex = vertexTmp;
  }
  free(graph);
  
  return;
}


/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 */
