/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 * chapter 01 - exercise 04 - solution 2
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#define MAX_LEN     (20+1)

// To debug the program set this to TRUE (not zero)
#define DEBUG       1

// Graph vertex V
typedef struct vertex_s {
  char *id;
  int nIn, nOut;
  int visited;
} vertex_t;

// Graph handle (wrapper)
typedef struct graph_s {
  int n;
  vertex_t *vertex;
  int **adj;
} graph_t;

// Solution structure
typedef struct sol_s {
  int *path;
  int *weight;
  int len;
  int sum;
} sol_t;

// Prototypes
graph_t *graph_init ();
graph_t *file_read (char *);
int vertex_find (graph_t *, char *);
void vertex_add (graph_t *, char *);
void graph_write (graph_t *);
void graph_regular (graph_t *);
void path_find (graph_t *, int, sol_t *);
void path_find_r (graph_t *, int, sol_t *, sol_t *);
void graph_dispose (graph_t *);

int main (int argc, char **argv) {
  graph_t *graph = NULL;
  int i;
  char id[MAX_LEN];
  sol_t best;

  if (argc != 2) {
    fprintf (stderr, "Run as: pgrmName inputFile\n");
    exit(EXIT_FAILURE);
  }

  graph = file_read (argv[1]);

#if DEBUG
  graph_write (graph);
#endif

  graph_regular (graph);

  best.len = best.sum = (-1);
  best.path = NULL;
  best.weight = NULL;
  do {
    fprintf (stdout, "Insert vertex_t name [end to quit] > ");
    scanf ("%s", id);

    if (strcmp (id, "end") != 0) {
      for (i=0; i<graph->n; i++) {
        graph->vertex[i].visited = 0;
      }
      i = vertex_find (graph, id);
      if (i == (-1)) {
        fprintf (stdout, "Invalid selection\n");
        continue;
      }
      path_find (graph, i, &best);
    }
  } while (strcmp (id, "end") != 0);

  free (best.path);
  free (best.weight);
  graph_dispose (graph);

  return 0;
}

/*
 * Read input file
 */
graph_t *file_read (char *name) {
  graph_t *graph;
  char name1[MAX_LEN], name2[MAX_LEN];
  int i, j, weight;
  FILE *fp;

  graph = graph_init ();

  // Read to create vertex array
  fp = fopen (name, "r");
  if (fp == NULL) {
    fprintf (stderr, "File open error\n");
    exit(EXIT_FAILURE);
  }
  
  while (fscanf(fp, "%s %d %s", name1, &weight, name2) != EOF) {
    i = vertex_find (graph, name1);
    if (i < 0) {
      vertex_add (graph, name1);
    }
    i = vertex_find (graph, name2);
    if (i < 0) {
      vertex_add (graph, name2);
    }
  }
  
  fclose (fp);

  // Allocate adjacency matrix
  graph->adj = malloc (graph->n * sizeof (int *));
  if (graph->adj==NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  
}  for (i=0; i<graph->n; i++) {
    // Use calloc to init elements to zero
    graph->adj[i] = calloc (graph->n, sizeof (int));
    if (graph->adj[i]==NULL) {
      fprintf (stderr, "Memory allocation error.\n");
      exit(EXIT_FAILURE);
    }
  }
  
  // Re-read
  fp = fopen (name, "r");
  if (fp == NULL) {
    fprintf (stderr, "File open error\n");
    exit(EXIT_FAILURE);
  }

  while (fscanf(fp, "%s %d %s", name1, &weight, name2) != EOF) {
    i = vertex_find (graph, name1);
    graph->vertex[i].nOut++;
    j = vertex_find (graph, name2);
    graph->vertex[j].nIn++;
    graph->adj[i][j] = weight;
  }

  fclose (fp);

  return (graph);
}

/*
 * Allocate graph handle and init fields
 */
graph_t *graph_init () {
  graph_t *g;

  g = (graph_t *) malloc (1 * sizeof(graph_t));
  if (g == NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }
  g->vertex = NULL;
  g->adj = NULL;
  g->n = 0;

  return g;
}

/*
 * Find vertex in vertex list
 */
int vertex_find (graph_t *graph, char *id) {
  int i;

  for (i=0; i<graph->n; i++) {
    if (strcmp (graph->vertex[i].id, id) == 0) {
      return i;
    }
  }

  return (-1);
}

/*
 * Add a vertex to the graph
 */
void vertex_add (graph_t *graph, char *id) {
  graph->n++;
  graph->vertex = realloc (graph->vertex, graph->n * sizeof (vertex_t));
  if (graph->vertex == NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit (EXIT_FAILURE);
  }
  graph->vertex[graph->n-1].id = strdup (id);
  graph->vertex[graph->n-1].nIn = 0;
  graph->vertex[graph->n-1].nOut = 0;
  graph->vertex[graph->n-1].visited = 0;

  return;
}

/*
 * Print-out graph
 */
void graph_write (graph_t *graph) {
  int i, j;

  fprintf (stdout, "GRAPH:\n");
  fprintf (stdout, "Num. of vertices: %d\n", graph->n);
  for (i=0; i<graph->n; i++) {
    fprintf (stdout, "#n=%d Vertex=%s inDegree=%d outDegree=%d\n",
      i, graph->vertex[i].id, graph->vertex[i].nIn, graph->vertex[i].nOut);
    for (j=0; j<graph->n; j++) {
      if (graph->adj[i][j] != 0) {
        fprintf (stdout, "    goes to #%s [weight = %d]\n",
	  graph->vertex[j].id, graph->adj[i][j]);
      }
    }
    fprintf (stdout, "\n");
  }

  return;
}

/*
 * Check regularity on a graph
 */
void graph_regular (graph_t *graph) {
  int i, in, out, reg;

  in = out = -1;
  reg = 1;
  for (i=0; i<graph->n; i++) {
    if (in == -1 && out == -1) {
      in = graph->vertex[i].nIn;
      out = graph->vertex[i].nOut;
      if (in != out) {
        reg = 0;
        break;
      }
    } else {
      if (in != graph->vertex[i].nIn || out != graph->vertex[i].nOut) {
        reg = 0;
        break;
      }
    }
  }

  if (reg != 0) {
    fprintf (stdout, "The given graph is regular\n\n");
  } else {
    fprintf (stdout, "The given graph is NOT regular\n\n");
  }

  return;
}

/*
 * Find a path
 */
void path_find (graph_t *graph, int n, sol_t *best) {
  sol_t curr;
  int i;

  curr.path = (int *) calloc (graph->n, sizeof (int));
  curr.weight = (int *) calloc (graph->n-1, sizeof (int));
  best->path = (int *) calloc (graph->n, sizeof (int));
  best->weight = (int *) calloc (graph->n-1, sizeof (int));
  if (curr.weight==NULL || curr.path==NULL ||
      best->weight==NULL || best->path==NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }
  curr.sum = curr.len = 0;

  path_find_r (graph, n, &curr, best);
  
  if (best->sum != -1) {
    for (i=0; i<=best->len; i++) {
      fprintf (stdout, "%s ", graph->vertex[best->path[i]].id);
      if (i != best->len)
        fprintf (stdout, "-<w=%d>- ", best->weight[i]);
    }
    fprintf (stdout, ": Total Weight=%d\n", best->sum);
    best->sum = -1;
  } else {
    fprintf (stdout, "No path found...\n");
  }

  free (curr.weight);
  free (curr.path);

  return;
}

/*
 * Find a path (recursive function)
 */
void path_find_r (
  graph_t *graph, int n, sol_t *curr, sol_t *best)
{
  int i;

  if (graph->vertex[n].visited == 1) {
    return;
  }

  curr->path[curr->len] = n;
  if (best->sum == -1 || best->sum<curr->sum) {
    for (i=0; i<=curr->len; i++) {
      best->path[i] = curr->path[i];
      if (i != curr->len)
        best->weight[i] = curr->weight[i];
    }
    best->sum = curr->sum;
    best->len = curr->len;
  }

  graph->vertex[n].visited = 1;

  for (i=0; i<graph->n; i++) {
    if (graph->adj[n][i] != 0) {
      curr->weight[curr->len] = graph->adj[n][i];
      curr->len = curr->len + 1;
      curr->sum = curr->sum + graph->adj[n][i];
      path_find_r (graph, i, curr, best);
      curr->len = curr->len - 1;
      curr->sum = curr->sum - graph->adj[n][i];
    }
  }

  graph->vertex[n].visited = 1;

  return;
}

/*
 * Free graph memory
 */
void graph_dispose (graph_t *graph) {
  int i;

  free (graph->vertex);
  for (i=0;  i<graph->n; i++) {
    free (graph->adj[i]);
  }
  free (graph->adj);
  free(graph);
  
  return;
}


/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 */
