/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 * chapter 01 - exercise 08 - solution 1
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define MAX (50+1)

typedef enum {M, F} sex;

// Main data structure
typedef struct node{
  char name[MAX];
  char surname[MAX];
  sex sex;
  int age;
  struct node *mother, *father;        // Parents
  struct node *next, *child;           // Right child, left sibling
  struct node *nextBySex, *nextByAge;  // Best paths
} node_t;

// Prototypes
node_t* file_read (char *, int *);
node_t* person_search (node_t *, char *, char *, int);
node_t* parent_assign (node_t *, node_t *, node_t *);
node_t* child_assign(node_t *, node_t *);
int path_find_by_age (node_t *, int);
int path_find_by_sex (node_t *, int, sex);
void bySex_write (node_t *);
void byAge_write (node_t *);

int main (int argc, char *argv[]) {
  node_t *pt, *pathbySex, *pathbyAge;
  int n_people, i, j, max;
  
  if (argc!=2) {
    fprintf (stderr, "Error in passing the arguments. Type <filename>");
    exit (EXIT_FAILURE);
  }

  /* Read input file */  
  pt = file_read (argv[1], &n_people);
  
  /* Find path with maximum number of males */
  for (max=0, i=0; i<n_people; i++){
    if (pt[i].sex == M) {
      j = path_find_by_sex (&pt[i], 0, M);
      if (j>max) {
        max = j;
        pathbySex = &pt[i];
      }
    }
  }
  fprintf (stdout, "Path by sex: %d males\n", max);
  bySex_write (pathbySex);

  /* Reset Links */
  for (i=0; i<n_people; i++){
    pt[i].nextBySex = NULL;
  }

  /* Find path with maximum number of females */
  for (max=0, i=0; i<n_people; i++){
    if (pt[i].sex == F) {
      j = path_find_by_sex (&pt[i], 0, F);
      if (j>max) {
        max = j;
        pathbySex = &pt[i];
      }
    }
  }
  fprintf (stdout, "Path by sex: %d females\n", max);
  bySex_write (pathbySex);

  /* Find path with maximum overall longevity */
  for (i=0, max=0; i<n_people; i++) {
    j = path_find_by_age (&pt[i], 0);
    if (j>max) {
      max=j;
      pathbyAge = &pt[i];
    }
  }
  fprintf (stdout, "Path by age: %d years\n", max);
  byAge_write (pathbyAge);
  
  free (pt);
  
  return 0;
}

/*
 *  Read file and create array of node_t elements
 */
node_t *file_read (char *filename, int *n_people){
  FILE *fp;
  char sex, childN[MAX], childS[MAX];
  char momN[MAX], momS[MAX], dadN[MAX], dadS[MAX];
  int i=0;
  node_t *pt, *current, *curr_mom, *curr_dad;
  
  fp = fopen (filename, "r");
  if (fp==NULL) {
    fprintf (stderr, "File open error.\n");
    exit (EXIT_FAILURE);
  }
  
  if (fscanf (fp, "%d", n_people) == EOF) {
    fprintf (stderr, "Error in reading file.");
    exit (EXIT_FAILURE);
  }
  
  pt = (node_t *) malloc ((*n_people) * sizeof (node_t));
  if (pt==NULL){
    fprintf (stderr, "Error in allocating the memory.");
    exit (EXIT_FAILURE);
  }
  
  while (i<*n_people){
    fscanf (fp, "%s %s %c %d", pt[i].name, pt[i].surname, &sex, &pt[i].age);

    pt[i].sex = (sex=='M') ? M : F;
    pt[i].mother = pt[i].father = pt[i].child = pt[i].next = NULL;
    pt[i].nextByAge = pt[i].nextBySex = NULL;

    i++;
  }
  
  while (fscanf (fp, "%s %s %s %s %s %s",
    childN, childS, dadN, dadS, momN, momS) != EOF) {
    current = person_search (pt, childN, childS, *n_people);
    curr_mom = person_search (pt, momN, momS, *n_people);
    curr_dad = person_search (pt, dadN, dadS, *n_people);
    current = parent_assign (current, curr_mom, curr_dad);
    curr_dad = child_assign (curr_dad, current);
    curr_mom = child_assign (curr_mom, current);
  }
  
  fclose (fp);
  
  return (pt);
}

/*
 *  Search a person within the array
 */
node_t *person_search (node_t* rp, char* name, char* surname, int n_people) {
  int i;
    
  for (i=0; i<n_people; i++){
    if (strcmp(rp[i].name, name)==0 &&
        strcmp(rp[i].surname, surname)==0) {
      return (&rp[i]);
    }
  }
    
  return NULL;
}

/*
 *  Set node pointers to given parents
 */
node_t *parent_assign (node_t* child, node_t* mom, node_t* dad) {
  child->mother = mom;
  child->father = dad;
    
  return (child);
}

/*
 *  Set a chld to a parent
 */
node_t *child_assign (node_t* parent, node_t* child) {
  child->next = parent->child;
  parent->child = child;
    
  return (parent);
}

/*
 *  Recursive search: Find the longest sequence of people with the same sex
 */ 
int path_find_by_sex (node_t *curr, int cnt, sex sex) {
  int cnt_new, cnt_max;
  node_t *pt;
  
  if (curr==NULL || curr->sex!=sex) {
    return (cnt);
  }

  cnt++;
  cnt_max = cnt;
  // Debug
  // printf ("visiting sex %s %s counter %d\n", pt->name, pt->surname, cnt); 
  
  pt = curr->child;
  while (pt != NULL){
    cnt_new = path_find_by_sex (pt, cnt, sex);

    if (cnt_new > cnt_max){
      cnt_max = cnt_new;
      (*curr).nextBySex = pt;
    }
   
    pt = pt->next;
  }
  
  return (cnt_max);
}

/*
 *  Recursive search: Find the longest sequence of people with max lifespan
 */ 
int path_find_by_age (node_t *curr, int age) {
  int age_new, age_max;
  node_t *pt;

  if (curr == NULL) {
    return age;
  }

  age = age + curr->age;
  age_max = age;
  // Debug
  // printf ("visiting age %s %s age %d\n", curr->name, curr->surname, age); 
  
  pt = curr->child;
  while (pt != NULL) {
    age_new = path_find_by_age (pt, age);

    if (age_new > age_max){
      age_max = age_new;
      (*curr).nextByAge = pt;
    }
      
    pt = pt->next;
  }
  
  return (age_max);
}

/*
 *  Write a tree-path following the nextBySex link
 */
void bySex_write (node_t* rp) {
  while (rp!=NULL) {
    fprintf (stdout, "%s %s\n", rp->name, rp->surname);
    rp=rp->nextBySex;
  }
    
  return;
}

/*
 *  Write a tree-path following the nextByAge link
 */
void byAge_write(node_t* rp) {
  while (rp!=NULL) {
    fprintf (stdout, "%s %s\n", rp->name, rp->surname);
    rp = rp->nextByAge;
  }
  
  return;
}


/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 */
