/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 * chapter 01 - exercise 10 - solution 1
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

void file_read (char *, int ***, int *);
void play (int **, int, int, int *, int);
void check (int **, int, int, int *);
void file_write (char *, int **, int, int);

int main (int argc, char *argv[]) {
  int i, nP, nH, sum;
  int **mat;

  nH = atoi (argv[3]);

  file_read (argv[1], &mat, &nP);

  sum = 0;
  play (mat, nP, nH, &sum, 0);

  fprintf (stdout, "Best solution with opt func = %d\n", sum);
  file_write (argv[2], mat, nP, nH);

  for (i=0; i<nH; i++) {
    free (mat[i]);
  }
  free (mat);

  return (0);
}

/*
 *  Allocate data structure and read file
 */
void file_read (char *name, int ***mat, int *n) {
  FILE *fp;
  int **tmp, i;

  fp = fopen (name, "r");
  if (fp==NULL) {
    fprintf (stderr, "File open error.\n");
    exit (EXIT_FAILURE);
  }

  if (fscanf (fp, "%d", n) == EOF) {
    fprintf (stderr, "File error.\n");
    return;
  }

  tmp = (int **) malloc (3 * sizeof (int *));
  if (tmp == NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    return;
  }
  for (i=0; i<3; i++) {
    tmp[i] = (int *) malloc (*n * sizeof (char *));
    if (tmp[i] == NULL) {
      fprintf (stderr, "Memory allocation error.\n");
      return;
    }
  }

  i = 0;
  while (fscanf (fp, "%d", &tmp[0][i]) != EOF && i<(*n)) {
    tmp[1][i] = 0;  // Storing Current Solution
    tmp[2][i] = 0;  // Storing Best Solution
    i++;
  }

  fclose (fp);
  *mat = tmp;

  return;
}

/*
 *  Recursive function
 */
void play (int **mat, int nP, int nH, int *sum, int level) {
  int i;

  if (level == nP) {
    check (mat, nP, nH, sum);
    return;
  }

  for (i=1; i<=nH; i++) {
    mat[1][level] = i;
    play (mat, nP, nH, sum, level+1);
  }

  return;
}

/*
 *  Check current solution to find the best one
 */
void check (int **mat, int nP, int nH, int *sum) {
  int i, j, *totA, tot;

  totA = (int *) malloc (nH * sizeof (int));
  if (totA == NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit (EXIT_FAILURE);
  }
  for (i=1; i<=nH; i++) {
    totA[i-1] = 0;
  }

  for (i=0; i<nP; i++) {
    totA[mat[1][i]-1] = totA[mat[1][i]-1] + mat[0][i];
  }

  tot = 0;
  for (i=1; i<=nH; i++) {
    for (j=i+1; j<=nH; j++) {
      tot = tot + abs (totA[i-1] - totA[j-1]);
    }
  }

  // New best: Copy current solution into best one
  if ((*sum==0) || (tot<(*sum))) {
    *sum = tot;
    for (i=0; i<nP; i++) {
      mat[2][i] = mat[1][i];
    }
  }
  
  free (totA);

  return;
}

/*
 *  Write output file
 */
void file_write (char *name, int **mat, int nP, int nH) {
  FILE *fp;
  int i, j;

  fp = fopen (name, "w");
  if (fp==NULL) {
    fprintf (stderr, "File open error.\n");
    exit (EXIT_FAILURE);
  }

  // For each heir ...
  for (i=1; i<=nH; i++) {
    // ... store his/her properties
    for (j=0; j<nP; j++) {
      if (mat[2][j] == i) {
        fprintf (fp, "%d ", mat[0][j]);
      }
    }
    fprintf (fp, "\n");
  }

  fclose (fp);

  return;
}

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 */
