/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 * chapter 01 - exercise 17 - solution 2
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include "impl2.h"

static void visitRow (struct rt *, float, int);
static struct rt *create (float, int, int, struct rt *);

void mat_insert (
  struct mt *m, int r, int c, float value
) {
  struct rt *rp1;

  rp1 = m->rp;

  // Corner case: The matrix is empty
  // OR a new row must be inserted before the first one
  if (rp1==NULL || r<rp1->r) {
    m->rp = create (value, r, c, rp1);

    return;
  }

  // Corner case: The correct row is the first one
  // Visit it
  if (rp1->r == r) {
    visitRow (rp1, value, c);
  }

  // Look-for the right row using two pointers
  while (rp1->rp!=NULL && rp1->rp->r<r) {
    rp1 = rp1->rp;
  }

  // If the current row rp1->rp is empty ... create a new row
  if (rp1->rp == NULL) {
    rp1->rp = create (value, r, c, rp1->rp);
    return;
  }
  // Otherwise ...
  if (rp1->rp->r == r) {
    // If rp1->rp is the correct row, visit it
    visitRow (rp1->rp, value, c);
  } else {
    // If rp1->rp goes after, insert a new row
    rp1->rp = create (value, r, c, rp1->rp);
  }

  return;
}

/*
 *  Create e new matrix row with a single element
 */
static struct rt *create (
  float value, int r, int c, struct rt *p
) {
  struct rt *rp;
  struct et *ep;

  rp = (struct rt *) malloc (sizeof (struct rt));
  ep = (struct et *) malloc (sizeof (struct et));
  if (rp==NULL || ep==NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit (EXIT_FAILURE);
  } 
  rp->r = r;
  rp->rp = p;
  rp->ep = ep;
  ep->c = c;
  ep->val = value;
  ep->ep = NULL;

  return (rp);
}

/*
 *  Visit a horizontal list (a matrix row)
 */
static void visitRow (struct rt *rp, float value, int c) {
  struct et *ep1, *ep2;

  ep1 = rp->ep;
  if (ep1==NULL || c<ep1->c) {
    ep2 = (struct et *) malloc (sizeof (struct et));
    if (ep2==NULL) {
      fprintf (stderr, "Memory allocation error.\n");
      exit (EXIT_FAILURE);
    } 
    ep2->c = c;
    ep2->val = value;
    ep2->ep = rp->ep;
    rp->ep = ep2;

    return;
  }

  if (ep1->c == c) {
    ep1->val = value;
  }

  while (ep1->ep!=NULL && ep1->ep->c<c) {
    ep1 = ep1->ep;
  }

  if (ep1->ep == NULL) {
    ep2 = (struct et *) malloc (sizeof (struct et));
    if (ep2==NULL) {
      fprintf (stderr, "Memory allocation error.\n");
      exit (EXIT_FAILURE);
    } 
    ep2->c = c;
    ep2->val = value;
    ep2->ep = NULL;

    ep1->ep = ep2;
  }
  if (ep1->ep->c == c) {
    ep1->val = value;
  } else {
    // Create and insert a new element
    ep2 = (struct et *) malloc (sizeof (struct et));
    if (ep2==NULL) {
      fprintf (stderr, "Memory allocation error.\n");
      exit (EXIT_FAILURE);
    } 
    ep2->c = c;
    ep2->val = value;
    ep2->ep = ep1->ep;

    ep1->ep = ep2;
  }

  return;
}

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 */
