/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 * chapter 01 - exercise 19 - solution 3
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdio.h>
#include <stdlib.h>

#define NS  4
#define NB  5

static int check (int [][NB], int, int, int, int []);

/*
 *  Version 1
 *  Powerset generated with arrangements with repetitions
 */
void switchAndBulb_ver1 (
  int mat[NS][NB], int ns, int nb, int *nBest, int switchBest[NS],
  int switchCurrent[NS], int level)
{
  int i, j, c, n, flag;

  // Termination
  if (level>=ns) {
    flag = 1;
    for (i=0; i<nb && flag==1; i++) {
      c = 0;
      for (j=0; j<ns; j++) {
        if (switchCurrent[j]==1) {
          c = c + mat[j][i];
        }
      }
      if ((c%2)==0) {
        flag = 0;
      }
    }
    if (flag==1) {
      n = 0;
      for (j=0; j<ns; j++) {
        n = n + switchCurrent[j];
      }
      if (n < *nBest) {
        *nBest = n;
        for (i=0; i<ns; i++) {
          switchBest[i] = switchCurrent[i];
        }
      }
    }
    return;
  }

  // Set All Switches
  switchCurrent[level] = 0;
  switchAndBulb_ver1 (mat, ns, nb, nBest, switchBest, switchCurrent, level+1);
  switchCurrent[level] = 1;
  switchAndBulb_ver1 (mat, ns, nb, nBest, switchBest, switchCurrent, level+1);

  return;
}

/*
 *  Version 2
 *  Powerset generated with simple n-choose-k combinations
 */
int switchAndBulb_ver2 (
  int mat[NS][NB], int ns, int nb, int switchCurrent[NS], int level,
  int k, int start
  )
{
  int i;

  // Termination
  if (level>=k) {
    if (check (mat, ns, nb, k, switchCurrent) == 1) {
      fprintf (stdout, "Solution found; ");
      fprintf (stdout, "%d Switches on -> ", k);
      for (i=0; i<k; i++)
	printf ("%d ", switchCurrent[i]);
      fprintf (stdout, "\n");
      return 1;
    } else {
      return 0;
    }
  }

  // Set All Switches
  for (i=start; i<ns; i++) {
    switchCurrent[level] = i;
    if (switchAndBulb_ver2 (mat, NS, NB, switchCurrent, level+1, k, i+1) == 1) {
      return 1;
    }
  }

  return 0;
}

static int check (int mat[NS][NB], int n, int m, int k, int switchCurrent[NS]) {
  int i, j, ok = 1, *bulbs;

  bulbs = calloc (m, sizeof (int));
  if (bulbs==NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }
  for (i=0; i<k; i++) {
    for (j=0; j<m; j++) {
      bulbs[j] += mat[switchCurrent[i]][j];
      bulbs[j] = bulbs[j]%2;
    }
  }
  for (i=0; i<m; i++) {
    ok &= bulbs[i];
  }
    
  free (bulbs);
  return ok;
}

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 */
