/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 * chapter 01 - exercise 21 - solution 3
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdio.h>
#include <stdlib.h>

#define C 8

static int cover_r (int [][C], int, int, int *, int, int);
static int check (int [][C], int *, int);

void cover (int m[][C], int n, int k) {
  int i, *selection;

  // Dynamically allocate the required array 
  selection = malloc (k * sizeof(int)); 
  if (selection==NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit (EXIT_FAILURE);
  }

  // Search solution
  if (cover_r (m, n, k, selection, 0, 0) == 1) {
    fprintf (stdout, "Cover of size=%d: ", k);
    for (i=0; i<k; i++) {
      fprintf (stdout, "%d ", selection[i]);
    }
    fprintf (stdout, "\n");
  } else {
    fprintf (stdout, "No Cover of size=%d.\n", k);
  }

  free (selection);

  return;
}

/*
 *  Generate simple combinations and check solution
 */
static int cover_r (
  int m[][C], int n, int k, int *selection, int level, int pos
) {
  int i;

  if (level==k) {
    if (check(m, selection, k)==1) {
      return 1;
    } else {
      return 0;
    }
  }

  for (i=pos; i<n; i++) {
    selection[level] = i;
    if (cover_r (m, n, k, selection, level+1, i+1)==1 ) {
      return 1;
    }
  }

  return 0;
}

/*
 *  Check whether a cover has been found
 */
static int check (int m[][C], int *selection, int k) {
  int i, r, c, *v;
 
  v = malloc (C * sizeof(int)); 
  if (v==NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit (EXIT_FAILURE);
  }
  for (i=0; i<C; i++) {
    v[i] = 0;
  }

  // Mark into v, select integers
  for (i=0; i<k; i++) {
    r = selection[i];
    for (c=0; c<C && m[r][c]!=0; c++) {
      v[m[r][c]-1]++;
    }
  }

  // Check for a complete cover
  for (i=0; i<C; i++) {
    if (v[i] == 0) {
      free (v);
      return 0;
    }
  }

  free (v);
  return 1;
}

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 */
