/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 * chapter 01 - exercise 24 - solution 1
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "util.h"
#include "stackPublic.h"
#include "queuePublic.h"

#define M (20+1) 

/* structure declaration */
typedef struct node {
  char *name;
  int level;
  struct node **children;
  int num;
} node_t;

/* function prototypes */
node_t *read_file (char *, int *);
node_t *create_node (char *name);
node_t *search_r (node_t *, char *);
void reverse_depth (node_t *, int);
char *max_path (node_t *root, char *strCurr, char *strBest);
void dispose_r (node_t *);

/*
 *  Main program
 */
int main (int argc, char *argv[]) {
  node_t *root;
  char *strBest;
  int size;

  util_check_m (argc>=2, "missing parameter(s)");  
  root = read_file(argv[1], &size);

  fprintf (stdout, "Nodes in reverse depth order:\n");
  reverse_depth (root, size);

  strBest = (char *) util_malloc (sizeof (char));
  strBest[0] = '\0';
  strBest = max_path (root, "", strBest);
  fprintf (stdout, "Max path string:\n   %s\n", strBest);
  free (strBest);

  dispose_r (root);
  return EXIT_SUCCESS;
}

/*
 *  Load the input file and create the tree
 */
node_t *read_file (char *filename, int *size_ptr) {
  node_t *tmp, *root=NULL;
  int i, n, cnt=0;
  char name[M];
  FILE *fp;
  
  fp = util_fopen (filename, "r");
  while (fscanf (fp, "%s %d", name, &n) != EOF) {
    if (root == NULL) {
      tmp = root = create_node (name);
    } else {
      tmp = search_r (root, name);
    }
    tmp->children = (node_t **) util_malloc (n * sizeof (node_t *));
    tmp->num = n;    
      
    for (i=0; i<n; i++) {
      fscanf (fp, "%s", name);
      tmp->children[i] = create_node (name);
    }
    cnt++;
  }  
  fclose (fp);
  *size_ptr = cnt;
  return root;
}

/*
 *  Allocate a new node and initialize it
 */
node_t *create_node (char *name) {
  node_t *ptr;
  
  ptr = (node_t *) util_malloc (sizeof (node_t));
  ptr->name = util_strdup (name);
  ptr->level = -1;
  ptr->children = NULL;
  ptr->num = 0;
  
  return ptr;
}

/*
 *  Recursively search a node in the tree given its name
 */
node_t *search_r (node_t *root, char *name) {
  node_t *tmp=NULL;
  int i;
  
  if (root != NULL) {
    if (strcmp (name, root->name) == 0) {
      tmp = root;
    } else {
      for (i=0; i<root->num && tmp==NULL; i++) {
        tmp = search_r (root->children[i], name);
      }
    }
  }
  
  return tmp;
}

/*
 *  Print all tree nodes following a decreasing depth sorting
 */
void reverse_depth (node_t *root, int size) {
  queue_t *qp;
  stack_t *sp;
  node_t *ptr;
  int i;

  /* Init queue and stack */
  qp = queue_init (size);
  sp = stack_init (size);

  /* Visit nodes from root following a FIFO order */
  root->level = 0;
  queue_put(qp, (void *) root);
  while (!queue_empty_m (qp)) {
    queue_get (qp, (void **)&ptr);
    for (i=0; i<ptr->num; i++) {
      ptr->children[i]->level = ptr->level+1;
      queue_put(qp, (void *)ptr->children[i]);
    }
    /* Insert visited noded in a LIFO stack */
    stack_push (sp, (void *)ptr);
  }

  /* Extract nodes from the stack to generate stdout */
  while (!stack_empty_m(sp)) {
    stack_pop(sp, (void **)&ptr);
    fprintf (stdout, "   %d %s\n", ptr->level, ptr->name);
  }

  /* Free queue and stack */
  queue_dispose (qp, NULL);
  stack_dispose (sp, NULL);
}

/*
 *  Recursively find the path with the maximum id length
 */
char *max_path (node_t *root, char *strCurr, char *strBest) {
  int i;
  char *tmp;
  
  /* To be precise, spaces to separate strings should be added... */
  tmp = (char *) util_malloc
    ((strlen( strCurr) + strlen (root->name)) * sizeof (char));
  sprintf (tmp, "%s%s", strCurr, root->name);

  if (root->num == 0) {
    /* Leaf reached; Update longest path string */
    if (strlen (tmp) > strlen (strBest)) {
      free (strBest);
      strBest = util_strdup (tmp);
    }
  } else {
    /* Visit all node children */
    for (i=0; i<root->num; i++) {
      /* Recur */
      strBest = max_path (root->children[i], tmp, strBest);
    }
  }
  free(tmp);
  return strBest;
}

/*
 *  Recursively deallocate the memory for the entire tree
 */
void dispose_r (node_t *root) {
  int i;
  
  if (root != NULL) {
    for (i=0; i<root->num; i++) {
      dispose_r (root->children[i]);
    }
    free (root->name);
    free (root);
  }
}

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 */
