/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 * chapter 01 - exercise 25 - solution 3
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdio.h>
#include <stdlib.h>

int matrix_swap (int **, int, int *, int, int, int);
int **malloc2d (int, int);
void free2d (int **, int);
int **file_read (char *, int *);
void matrix_write (int **, int);
  
int main (int argc, char *argv[]) {
  int n, N, k, found;
  int **mat, *sol;
  
  mat = file_read (argv[1], &n);
  found = 0;
  N = 2*n;
  sol = calloc (N, sizeof (int));
  if (sol==NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit (EXIT_FAILURE);
  }
  
  fprintf (stdout, "Initial Matrix:\n");
  matrix_write (mat, n);

  for (k=1; k<=N && found==0; k++) {
    if (matrix_swap (mat, N, sol, 0, k, 0)) {
      found = 1;
    }
  }
  if (found==0) {
    fprintf (stdout, "No solution found.\n");
  }

  free (sol);
  free2d (mat, n);

  return 0;
}

int **file_read (char *name, int *n) {
  int i, j, **mat;
  FILE *fp;

  fp = fopen (name, "r");
  if (fp==NULL) {
    fprintf (stderr, "File open error.\n");
    exit(EXIT_FAILURE);
  }
  fscanf (fp, "%d", n);

  mat = malloc2d (*n, *n);

  for (i=0; i<*n; i++) {
    for (j=0; j<*n; j++) {
      fscanf(fp, "%d", &mat[i][j]);
    }
  }

  fclose (fp);

  return mat;
}

void matrix_write (int **m, int n) {
  int i, j;

  for (i=0; i<n; i++) {
    for (j=0; j<n; j++) {
      fprintf (stdout, "%d ", m[i][j]);
    }
    fprintf (stdout, "\n");
  }

  return;
}

int **malloc2d (int r, int c) {
  int i;
  int **mat;

  mat = (int **) malloc (r * sizeof (int *));
  if (mat == NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }
  for (i=0; i<r; i++) {
    mat[i] = (int *) malloc(c * sizeof (int));
    if (mat[i]==NULL) {
      fprintf (stderr, "Memory allocation error.\n");
      exit(EXIT_FAILURE);
    }
  }

  return mat;
}

void free2d (int **mat, int r) {
  int i;

  for (i=0; i<r; i++) {
    free (mat[i]);
  }
  free (mat);

  return;
}

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 */
