/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 * chapter 01 - exercise 26 - solution 1
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "symbolTable.h"

typedef struct st_item_s {
  char *id;
} st_item_t;

struct st_s {
  st_item_t *items;
  int n, m;
};

st_t st_init (int maxN) {
  st_t table;
  
  table = calloc (1, sizeof (struct st_s));
  if (table == NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit (EXIT_FAILURE);
  }
  table->n = 0;
  table->m = maxN;
  table->items = calloc (table->m, sizeof (st_item_t));
  if (table->items == NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit (EXIT_FAILURE);
  }

  return table;
}

int st_get_size (st_t table) {
  return table->n;
}

int st_add (st_t table, char *str) {
  if (table->n == table->m) {
    table->m *= 2;
    table->items = realloc (table->items, table->m*sizeof(st_item_t));
    if (table->items == NULL) {
      fprintf (stderr, "Memory allocation error.\n");
      exit (EXIT_FAILURE);
    }
  }
  table->items[table->n++].id = strdup (str);

  return table->n-1;
}

int st_find(st_t table, char *str) {
  int i;

  for (i=0; i<table->n; i++) {
    if (!strcmp(table->items[i].id, str)) {
      return i;
    }
  }
  
  return -1;
}

void st_dispose (st_t table) {
  int i;

  if (table == NULL)
    return;
  if (table->items != NULL) {
    for (i=0; i<table->n; i++) {
      if (table->items[i].id != NULL) {
	free (table->items[i].id);
      }
    }
    free (table->items);
  }
  free (table);

  return;
}

char *st_read(st_t table, int index) {
  if (index < 0 || index > table->n)
    return NULL;

  return (table->items[index].id);
}

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 */
