/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 * chapter 01 - exercise 27 - solution 3
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdlib.h>
#include <stdio.h>

static int powerset_ver1 (int **, int, int, int, int *, int);
static int powerset_ver2 (int **, int, int, int *);
static int powerset_r_ver2 (int **, int, int, int *, int, int, int);
static int check_ver1 (int **, int *, int);
static int check_ver2 (int **, int *, int, int);

void partition (int **mat, int n, int k) {
  int i, *sol;

  sol = malloc (n* sizeof (int));
  if (sol==NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit (EXIT_FAILURE);
  }

  // Call version 1
  if (powerset_ver1 (mat, n, k, 0, sol, 0) == 1) {
    fprintf (stdout, "Solution Version 1: ");
    for (i=0; i<n; i++) {
      fprintf (stdout, "%d ", sol[i]);
    }
    fprintf (stdout, "\n");  
  } else {
    fprintf (stdout, "No solution.\n");
  }

  // Call version 2
  if (powerset_ver2 (mat, n, k, sol) == 0) {
    fprintf (stdout, "No solution.\n");
  }

  free (sol);

  return;
}

/* Version 1 */
static int powerset_ver1 (
  int **mat, int n, int k, int pos, int *sol, int moved
) {
  if (pos >= n) {
    if (moved < k || !check_ver1(mat, sol, n))
      return 0;
    else
      return 1;  
  }

  sol[pos] = 0;
  if (powerset_ver1 (mat, n, k, pos+1, sol, moved))
    return 1;
  sol[pos] = 1;
  if (powerset_ver1 (mat, n, k, pos+1, sol, moved+1))
    return 1;

  return 0;
}

static int check_ver1 (int **mat, int *sol, int n) {
  int i, j;

  for (i=0; i<n-1; i++) {
    for (j=i+1; j<n; j++)
      if (mat[i][j] && sol[i] != sol[j])
        return 0;  
  }
  return 1;
}

/* Version 2 */
static int powerset_ver2 (
  int **mat, int n, int k, int *sol
) {
  int i, j;

  for (i=k; i<=n; i++) {
    if (powerset_r_ver2 (mat, n, k, sol, i, 0, 0) == 1) {
      fprintf (stdout, "Solution Version 2: ");
      for (j=0; j<i; j++) {
        fprintf (stdout, "%d ", sol[j]);
      }
      fprintf (stdout, "\n");
      return 1;  
    } 
  }

  return 0;
}

static int powerset_r_ver2 (
  int **mat, int n, int k, int *sol,
  int stop, int pos, int start
) {
  int i;

  if (pos >= stop) {
    if (check_ver2(mat, sol, n, pos))
      return 1;
    else
      return 0;  
  }

  for (i=start; i<n; i++) {
    sol[pos] = i;
    if (powerset_r_ver2 (mat, n, k, sol, stop, pos+1, i+1) == 1)
      return 1;
  }

  return 0;
}

static int check_ver2 (int **mat, int *sol, int n, int pos) {
  int i, j, k, yes;

  for (k=pos, i=0; i<n; i++) {
    for (yes=1, j=0; j<pos; j++) {
      if (sol[j]==i) {
        yes = 0;
      }
    }
    if (yes==1) {
      sol[k] = i;
      k++;
    }
  }

  for (i=0; i<pos; i++) {
    for (j=pos; j<n; j++) {
      if (mat[sol[i]][sol[j]]==1)
        return 0;
    }
  }

  return (1);
}

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 */
