/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 * chapter 01 - exercise 28 - solution 1
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <stdio.h>

#define MAXS 21
#define PLUS_INF 1e6

// Reference from products to offers
typedef struct node_s {
  int offer_n;
  struct node_s *next;
} node_t;

// A single product
typedef struct product_s {
  char name[MAXS];
  int price;
  node_t *offer_list; 
} product_t;

// A single offer
typedef struct offer_s {
  char name[MAXS];
  int price;
  int *pro;
  int *qta;
  int n;
} offer_t;

// Array of products
typedef struct products_s {
  product_t *array;
  int n;
} products_t;

// Array of offers
typedef struct offers_s {
  offer_t *array;
  int n;
} offers_t;

// Prototypes
void product_read (char *, products_t *, offers_t *);
int product_find (products_t, char *);
void offer_add (product_t *, int);
void offer_per_product_write (char *, products_t, offers_t);
void common_product_write (char *, char *, products_t, offers_t);
int offer_find (offers_t, char *);
int *list_read (char *, products_t);
void mis_cost_grocery (products_t, offers_t, int *);
void powerset (int, int *, int *, int *, int, products_t, offers_t, int *);
void check (int *, int *, int *, int, products_t, offers_t, int *);

int main (int argc, char **argv) {
  products_t product;
  offers_t offer;
  int *grocery;

  fprintf (stdout, "Reading ...\n");
  product_read (argv[1], &product, &offer);
  
  // Problem 1: Bargains with same product
  fprintf (stdout, "Offer per product ...\n");
  offer_per_product_write ("Nutella", product, offer);

  // Problem 2: Given 2 bargains find common products
  fprintf (stdout, "Common products ...\n");
  common_product_write ("OFF1", "OFF4", product, offer);

  // Problem 3: Find minimum cost
  fprintf (stdout, "Best grocery ...\n");
  grocery = list_read (argv[2], product);
  mis_cost_grocery(product, offer, grocery);
  
  return 0;
}

/*
 *  Read all products and all bargain offers
 */
void product_read (char *name, products_t *product, offers_t *offer) {
  FILE *fp;
  int i, j, index;
  char str[MAXS];

  fp = fopen (name, "r");
  if (fp == NULL) {
    fprintf (stderr, "File open error.\n");
    exit (EXIT_FAILURE);
  }

  fscanf (fp, "%d %d", &product->n, &offer->n);
  product->array = calloc (product->n, sizeof (product_t));
  if (product->array==NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit (EXIT_FAILURE);
  }

  for (i=0; i<product->n; i++) {
    fscanf (fp, "%s %d", product->array[i].name, &product->array[i].price);
  }

  offer->array = calloc (offer->n, sizeof (offer_t));
  if (offer->array==NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit (EXIT_FAILURE);
  }
  for (i=0; i<offer->n; i++) {
    fscanf (fp, "%s %d %d",
      offer->array[i].name, &offer->array[i].price, &offer->array[i].n);
    offer->array[i].pro = calloc (offer->array[i].n, sizeof (int));
    offer->array[i].qta = calloc (offer->array[i].n, sizeof (int));
    if (offer->array[i].pro==NULL || offer->array[i].qta==NULL) {
      fprintf (stderr, "Memory allocation error.\n");
      exit (EXIT_FAILURE);
    }
    for (j=0; j<offer->array[i].n; j++) {
      fscanf (fp, "%s %d", str, &offer->array[i].qta[j]);
      index = product_find (*product, str);
      assert (index != -1);
      offer->array[i].pro[j] = index;
      offer_add (&(product->array[index]), j);
    }
  }

  fclose (fp);

  return;
}

/*
 * Find a product within the array of products
 */
int product_find (products_t product, char *name) {
  int i;

  for (i=0; i<product.n; i++) {
    if (!strcmp (product.array[i].name, name))
      return i;
  }
  return -1;
}

/*
 * Add a bargain offer reference to a product
 */
void offer_add (product_t *p, int off) {
  node_t *node;

  node = calloc (1, sizeof (node_t));
  if (node==NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit (EXIT_FAILURE);
  }

  node->offer_n = off;
  node->next = p->offer_list;
  p->offer_list = node;

  return;
}

/*
 *  Given a product find all bargain offers including it
 */
void offer_per_product_write (char *name, products_t product, offers_t offer) {
  node_t *tmp;
  int p;

  p = product_find (product, name);
  if (p == -1)
    return;

  tmp = product.array[p].offer_list;
  while (tmp != NULL) {
    if (offer.n>tmp->offer_n && tmp->offer_n>=0)
      fprintf (stdout, "%s\n", offer.array[tmp->offer_n].name); 
    tmp = tmp->next;
  }

  return;
}

/*
 *  Given two bargain offers find common products
 */
void common_product_write (
  char *name1, char *name2, products_t product, offers_t offer
) {
  int i, o1, o2, *mark;

  o1 = offer_find (offer, name1);
  o2 = offer_find (offer, name2);
  if (o1==(-1) || o2==(-1)) {
    return;
  }

  mark = calloc (product.n, sizeof (int));
  if (mark==NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit (EXIT_FAILURE);
  }

  for (i=0; i<offer.array[o1].n; i++) {
    mark[offer.array[o1].pro[i]]++;
  } 
  for (i=0; i<offer.array[o2].n; i++) {
    mark[offer.array[o2].pro[i]]++; 
  } 
  for (i=0; i<product.n; i++) {
    if (mark[i] == 2) {
      if (product.n>i && i>=0) {
        fprintf (stdout, "%s\n", product.array[i].name); 
      }
    }
  }

  free (mark);
}

/*
 * Find a bargain offer within the array of bargain offers
 */
int offer_find (offers_t offer, char *name) {
  int i;

  for (i=0; i<offer.n; i++) {
    if (!strcmp (offer.array[i].name, name))
      return i;
  }

  return -1;
}

/*
 *  Read the grocery list from file
 */
int *list_read (char *name, products_t product) {
  FILE *fp;
  int *grocery, i, qta, id, ns;
  char str[MAXS];

  grocery = calloc (product.n, sizeof (int));
  if (grocery==NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit (EXIT_FAILURE);
  }

  fp = fopen (name, "r");
  if (fp == NULL) {
    fprintf (stderr, "File open error.\n");
    exit (EXIT_FAILURE);
  }

  fscanf (fp, "%d", &ns);
  for (i=0; i<ns; i++) {
    fscanf (fp, "%s %d", str, &qta);
    id = product_find (product, str);
    if (id != -1)
      grocery[id] = qta;
  }

  fclose(fp);

  return grocery;
}

/*
 *  Find the optimum set of purchases
 */
void mis_cost_grocery (products_t product, offers_t offer, int *grocery) {
  int bestprice = PLUS_INF;
  int *sol, *best_sol;

  sol = calloc (offer.n, sizeof (int));
  best_sol = calloc (offer.n, sizeof (int));
  if (sol==NULL || best_sol==NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }

  powerset (0, sol, best_sol, &bestprice, 0, product, offer, grocery); 

  fprintf (stdout, "Total cost: %d\n", bestprice);
  
  free (sol);
  free (best_sol);

  return;
}

/*
 * Generate the powerset
 */
void powerset (
  int pos, int *sol, int *best_sol, int *bestp,
  int currp, products_t product, offers_t offer, int *grocery
) {
  if (pos == offer.n) {
    check(sol, best_sol, bestp, currp, product, offer, grocery);
    return;  
  }
  
  if ((currp + offer.array[pos].price) <= *bestp) {
    sol[pos] = 1;
    powerset (pos+1, sol, best_sol, bestp, currp + offer.array[pos].price,
      product, offer, grocery);
  }
  sol[pos] = 0;
  powerset (pos+1, sol, best_sol, bestp, currp, product, offer, grocery);

  return;
}

/*
 * Check current grocery list
 */
void check (
  int *sol, int *best_sol, int *bestp, int currp, products_t product,
  offers_t offer, int *grocery
) {
  int i, j, *current_grocery;

  current_grocery = calloc(product.n, sizeof(int));
  if (current_grocery==NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }

  // Create current grocery list
  for (i=0; i<offer.n; i++) {
    if (sol[i] != 0) {
      for (j=0; j<offer.array[i].n; j++) {
        current_grocery[offer.array[i].pro[j]] += offer.array[i].qta[j];
      }
    }
  }

  // Find missing items
  for (i=0; i<product.n; i++) {
    if (current_grocery[i] < grocery[i]) {
      currp += ((grocery[i]-current_grocery[i]) * product.array[i].price);
    }
  }

  if (currp < *bestp) {
    *bestp = currp;
    for (i=0; i<offer.n; i++) {
      best_sol[i] = sol[i];
    }
  }

  free (current_grocery);

  return;
}




/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 */
